#include "hxd.h"
#include <string.h>
#include <unistd.h>
#include <netinet/in.h>
#include <sys/socket.h>
#include <errno.h>
#include "xmalloc.h"

static void
htxf_close (int fd)
{
	char buf[16];
	struct htxf_conn *htxf = hxd_files[fd].conn.htxf;

	close(fd);
	FD_CLR(fd, &hxd_rfds);
	FD_CLR(fd, &hxd_wfds);
	inet_ntoa_r(htxf->sockaddr.sin_addr, buf, sizeof(buf));
	hxd_log("%s:%u -- htxf connection closed", buf, ntohs(htxf->sockaddr.sin_port));
	if (htxf->in.buf)
		xfree(htxf->in.buf);
	xfree(htxf);
	memset(&hxd_files[fd], 0, sizeof(struct hxd_file));
	if (high_fd == fd) {
		for (fd--; fd && !FD_ISSET(fd, &hxd_rfds); fd--)
			;
		high_fd = fd;
	}
}

static void
htxf_read_hdr (int fd)
{
	unsigned int i;
	ssize_t r;
	struct htxf_conn *htxf = hxd_files[fd].conn.htxf;

	r = read(fd, &htxf->in.buf[htxf->in.pos], htxf->in.len);
	if (r == 0 || (r < 0 && errno != EWOULDBLOCK && errno != EINTR)) {
		htxf_close(fd);
	} else {
		htxf->in.pos += r;
		htxf->in.len -= r;
		if (!htxf->in.len) {
			struct htxf_hdr *h = (struct htxf_hdr *)htxf->in.buf;
			u_int32_t ref;
			struct htlc_conn *htlcp;

			if (ntohl(h->magic) != HTXF_MAGIC_INT) {
				htxf_close(fd);
				return;
			}
			ref = ntohl(h->ref);
			for (htlcp = htlc_list->next; htlcp; htlcp = htlcp->next) {
				for (i = 0; i < HTXF_GET_MAX; i++) {
					if (htlcp->htxf_out[i] && htlcp->htxf_out[i]->ref == ref) {
					}
				}
				for (i = 0; i < HTXF_PUT_MAX; i++) {
				   	if (htlcp->htxf_in[i] && htlcp->htxf_in[i]->ref == ref) {
					}
				}
			}
			htxf_close(fd);
		}
	}
}

static void
htxf_timeout (struct htxf_conn *htxf)
{
	htxf_close(htxf->fd);
}

static void
htxf_listen_ready_read (int fd)
{
	int s;
	struct sockaddr_in saddr;
	int siz = sizeof(saddr);
	char buf[16];
	struct htxf_conn *htxf;

	s = accept(fd, (struct sockaddr *)&saddr, &siz);
	if (s < 0) {
		hxd_log("htxf: accept: %s", strerror(errno));
		return;
	}
	if (s >= hxd_open_max) {
		hxd_log("%s:%d: %d >= hxd_open_max (%d)", __FILE__, __LINE__, s, hxd_open_max);
		close(s);
		return;
	}
	fd_closeonexec(s, 1);
	fd_blocking(s, 0);
	inet_ntoa_r(saddr.sin_addr, buf, sizeof(buf));
	hxd_log("%s:%u -- htxf connection accepted", buf, ntohs(saddr.sin_port));

	htxf = xmalloc(sizeof(struct htxf_conn));
	memset(htxf, 0, sizeof(struct htxf_conn));
	htxf->fd = s;
	htxf->sockaddr = saddr;
	qbuf_set(&htxf->in, 0, SIZEOF_HTXF_HDR);

	hxd_files[s].ready_read = htxf_read_hdr;
	hxd_files[s].ready_write = 0;
	hxd_files[s].conn.htxf = htxf;

	timer_add_secs(10, htxf_timeout, htxf);
}

void
htxf_init (void)
{
	struct sockaddr_in saddr;
	int x;
	int htxf_listen_sock;

	htxf_listen_sock = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP);
	if (htxf_listen_sock < 0) {
		hxd_log("%s:%d: socket: %s", __FILE__, __LINE__, strerror(errno));
		exit(1);
	}
	if (htxf_listen_sock >= hxd_open_max) {
		hxd_log("%s:%d: %d >= hxd_open_max (%d)", __FILE__, __LINE__, htxf_listen_sock, hxd_open_max);
		close(htxf_listen_sock);
		exit(1);
	}
	x = 1;
	setsockopt(htxf_listen_sock, SOL_SOCKET, SO_REUSEADDR, &x, sizeof(x));
	saddr.sin_family = AF_INET;
	saddr.sin_port = htons(hxd_cfg.options.htls_port + 1);

	if (bind(htxf_listen_sock, (struct sockaddr *)&saddr, sizeof(saddr)) < 0) {
		hxd_log("%s:%d: bind: %s", __FILE__, __LINE__, strerror(errno));
		exit(1);
	}
	if (listen(htxf_listen_sock, 5) < 0) {
		hxd_log("%s:%d: listen: %s", __FILE__, __LINE__, strerror(errno));
		exit(1);
	}

	hxd_files[htxf_listen_sock].ready_read = htxf_listen_ready_read;
	FD_SET(htxf_listen_sock, &hxd_rfds);
	if (high_fd < htxf_listen_sock)
		high_fd = htxf_listen_sock;
	fd_closeonexec(htxf_listen_sock, 1);
	fd_blocking(htxf_listen_sock, 0);
}
