#include <stdio.h>
#include <stdarg.h>
#include <stdlib.h>
#include <string.h>
#include <termios.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <errno.h>
#include <signal.h>
#include <sys/ioctl.h>
#include "hxd.h"
#include "hx.h"
#include "xmalloc.h"
#include "getopt.h"
#ifdef HAVE_TERMCAP_H
#include <termcap.h>
#if !defined(__NetBSD__)
#define HAVE_TERMCAP_PC_BC_UP 1
#endif
#endif

#ifndef HAVE_TCGETATTR
#define tcgetattr(fd,tty)
#endif
#ifndef HAVE_TCSETATTR
#define tcsetattr(fd,opt,tty)
#endif

#define tty_fd 0

int LI = 24, CO = 80;
char term_name[256] = "vt100", *HO, *ME, *MR, *US, *CL, *CE, *CM, *CS, *SF, *SR, *DC, *IC;
static char termcap[2048], tbuf[2048];
static struct termios old_tty, new_tty;
static int dumb = 0;

static char term_flush_buf[32];
static int term_flush_len = 0;

static void
term_flush (void)
{
	write(1, term_flush_buf, term_flush_len);
	term_flush_len = 0;
}

static int
term_putc (int c)
{
	if (term_flush_len == sizeof(term_flush_buf)) {
		term_flush();
		term_flush_len = 0;
	}
	term_flush_buf[term_flush_len] = (char)c;
	term_flush_len++;

	return c;
}

static void
term_reset (void)
{
	if (dumb)
		return;
	tcsetattr(tty_fd, TCSADRAIN, &old_tty);
	tputs(tgoto(CS, -1, 0), 0, term_putc);
	tputs(tgoto(CM, 0, LI), 0, term_putc);
	term_flush();
}

static void
term_goto_line (int line)
{
	if (dumb)
		return;
	tputs(tgoto(CM, rl_point % CO, line), 0, term_putc);
	term_flush();
}

static void
term_ce (void)
{
	if (dumb)
		return;
	tputs(CE, 0, term_putc);
	term_flush();
}

static void
term_mode_clear (void)
{
	if (dumb)
		return;
	tputs(ME, 0, term_putc);
	term_flush();
}

static void
term_mode_underline (void)
{
	if (dumb)
		return;
	tputs(US, 0, term_putc);
	term_flush();
}

static void term_status (void);

static char current_time_str[16];

static void
term_clock_timer (char *time_str)
{
	time_t t;
	struct tm tm;

	time(&t);
	if (localtime_r(&t, &tm))
		strftime(time_str, sizeof(current_time_str), "%H:%M", &tm);
	term_status();

	timer_add_secs(60 - tm.tm_sec, term_clock_timer, time_str);
}

static void
term_status (void)
{
	int len;
	char buf[256], addr[16];
	char *col;

	if (dumb)
		return;
	col = hx_htlc.color % 4 > 1 ? (hx_htlc.color % 2 ? RED : RED_BOLD)
				    : (hx_htlc.color % 2 ? GREEN : GREEN_BOLD);
	tputs(tgoto(CM, 0, LI - 3), 0, term_putc);
	term_flush();
	inet_ntoa_r(hx_htlc.sockaddr.sin_addr, addr, sizeof(addr));
	len = sprintf(buf, "[ %s ... hx version %s]\n[%s%s%s<%u>(%u) @ %s:%u]", current_time_str, version,
		col, hx_htlc.name, DEFAULT, hx_htlc.icon, hx_htlc.uid, addr, ntohs(hx_htlc.sockaddr.sin_port));
	write(1, buf, len);
	tputs(CE, 0, term_putc);
	tputs(tgoto(CM, rl_point % CO, LI - 1), 0, term_putc);
	term_flush();
}

static RETSIGTYPE
term_sigwinch (int sig __attribute__((__unused__)))
{
	struct winsize win_size;
	char LINES[16], COLUMNS[16];

	if (ioctl(1, TIOCGWINSZ, &win_size))
		goto ret;

	LI = win_size.ws_row;
	CO = win_size.ws_col;
#if defined(HAVE_PUTENV)
	sprintf(LINES, "LINES=%.4u", win_size.ws_row);
	putenv(LINES);
	sprintf(COLUMNS, "COLUMNS=%.4u", win_size.ws_col);
	putenv(COLUMNS);
#endif

	tputs(tgoto(CS, LI - 4, 0), 0, term_putc);
	term_flush();
ret:
	{}
}

static RETSIGTYPE
term_sigtstp (int sig __attribute__((__unused__)))
{
	tcgetattr(tty_fd, &new_tty);
	term_reset();
	write(1, "\r", 1);
	raise(SIGSTOP);
}

extern RETSIGTYPE sig_chld (int sig);

static RETSIGTYPE
term_sigcont (int sig __attribute__((__unused__)))
{
	tcsetattr(tty_fd, TCSADRAIN, &new_tty);
	tputs(tgoto(CS, LI - 4, 0), 0, term_putc);
	term_status();
	tputs(tgoto(CM, 0, LI - 1), 0, term_putc);
	term_flush();
	rl_on_new_line();
	rl_redisplay();
	sig_chld(0);
}

static void
term_install_signal_handlers (void)
{
	struct sigaction act;

	act.sa_flags = 0;
	sigfillset(&act.sa_mask);

	act.sa_handler = term_sigcont;
	sigaction(SIGINT, &act, 0);
	sigaction(SIGCONT, &act, 0);

	act.sa_handler = term_sigtstp;
	sigaction(SIGTSTP, &act, 0);

	act.sa_handler = term_sigwinch;
	sigaction(SIGWINCH, &act, 0);
}

static int redraw();
static void line_up(), line_down(), page_up(), page_down(), home(), end();

static void
term_rl_init (void)
{
	Keymap kmap = rl_get_keymap();

	rl_bind_key(0xc, redraw);
	rl_generic_bind(ISFUNC, "\033[1~", (char *)home, kmap);
	rl_generic_bind(ISFUNC, "\033[2~", (char *)line_up, kmap);
	rl_generic_bind(ISFUNC, "\033[3~", (char *)line_down, kmap);
	rl_generic_bind(ISFUNC, "\033[4~", (char *)end, kmap);
	rl_generic_bind(ISFUNC, "\033[5~", (char *)page_up, kmap);
	rl_generic_bind(ISFUNC, "\033[6~", (char *)page_down, kmap);
}

static int
term_init ()
{
	char *ep, *bp = tbuf;
	int i;

	if ((ep = getenv("TERM"))) {
		strncpy(term_name, ep, sizeof(term_name) - 1);
		term_name[sizeof(term_name) - 1] = 0;
	}
	if (dumb)
		goto want_dumb;
	i = tgetent(termcap, term_name);
	if (i <= 0) {
		if (i == -1) {
			write(1, "tgetent had trouble accessing the termcap database\n", 51);
		} else {
			write(1, "tgetent did not find an entry for '", 25);
			write(1, term_name, strlen(term_name));
			write(1, "'\n", 2);
		}
want_dumb:
		write(1, "running in dumb terminal mode\n", 31);
		goto ret;
	}
	if (!(ep = getenv("COLUMNS")) || !(CO = strtol(ep, 0, 0)))
		if ((CO = tgetnum("co")) == -1)
			CO = 80;
	if (!(ep = getenv("LINES")) || !(LI = strtol(ep, 0, 0)))
		if ((LI = tgetnum("li")) == -1)
			LI = 24;

	HO = tgetstr("ho", &bp);
	ME = tgetstr("me", &bp);
	MR = tgetstr("mr", &bp);
	US = tgetstr("us", &bp);
	CL = tgetstr("cl", &bp);
	CM = tgetstr("cm", &bp);
	CE = tgetstr("ce", &bp);
	CS = tgetstr("cs", &bp);
	DC = tgetstr("dc", &bp);
	IC = tgetstr("ic", &bp);
	SF = tgetstr("sf", &bp);
	SR = tgetstr("sr", &bp);

#ifdef HAVE_TERMCAP_PC_BC_UP
	ep = tgetstr("pc", &bp);
	PC = ep ? *ep : 0;
	BC = tgetstr("le", &bp);
	UP = tgetstr("up", &bp);
#endif

	tcgetattr(tty_fd, &old_tty);
	new_tty = old_tty;
	new_tty.c_cc[VMIN] = 1;
	new_tty.c_cc[VTIME] = 0;
	new_tty.c_lflag &= ~ICANON;
	new_tty.c_iflag &= ~IXON;
	tcsetattr(tty_fd, TCSADRAIN, &new_tty);

	tputs(tgoto(CS, LI - 4, 0), 0, term_putc);
	tputs(CL, 0, term_putc);
	term_clock_timer(current_time_str);
	term_status();

ret:
	term_install_signal_handlers();

	term_rl_init();

	return tty_fd;
}

struct screen {
	char *buf;
	size_t size;
	int li, page_lines;
};

static struct screen __scr = {0, 0, 0, 0}, *scr_front = &__scr;

static int last_co = 0;

static void
term_puts (char *buf, size_t len)
{
	if (!dumb) {
		tputs(tgoto(CM, last_co, LI - 4), 0, term_putc);
		if (!last_co)
			term_putc('\n');
		term_flush();
		if (buf[len - 1] == '\n') {
			last_co = 0;
			len--;
		} else last_co += len;
	}
	write(1, buf, len);
	if (!dumb) {
		tputs(tgoto(CM, rl_point % CO, LI - 1), 0, term_putc);
		term_flush();
	}
}

static int
lines (char *buf, size_t size)
{
	char *p, *end = buf + size;
	int l = 0;

	for (p = buf; p < end; p++) {
		if (*p == '\n')
			l++;
	}

	return l;
}

static char *
line (struct screen *scr, int line)
{
	char *p, *end = scr->buf + scr->size;
	int l = 0;

	if (!line)
		return scr->buf;
	for (p = scr->buf; p < end; p++) {
		if (*p == '\n') {
			l++;
			if (l == line)
				return (p + 1) >= end ? 0 : (p + 1);
		}
	}
	if (l == line - 1)
		return p;

	return 0;
}

static void
draw (char *ln0, size_t len)
{
	tputs(CL, 0, term_putc);
	term_status();
	tputs(tgoto(CM, 0, 0), 0, term_putc);
	term_flush();
	write(1, ln0, len);
	tputs(tgoto(CM, 0, LI - 1), 0, term_putc);
	term_flush();
	rl_on_new_line();
	rl_redisplay();
}

static int
redraw ()
{
	struct screen *scr = scr_front;
	char *ln0 = 0, *ln1 = 0;

	if (!scr->size)
		goto cl;
	if (scr->li > scr->page_lines)
		ln0 = line(scr, (scr->li - scr->page_lines));
	else
		ln0 = scr->buf;
	ln1 = line(scr, scr->li + 1);
	if (!ln1)
		ln1 = scr->buf + scr->size;

cl:
	tputs(CL, 0, term_putc);
	term_status();
	tputs(tgoto(CM, 0, 0), 0, term_putc);
	term_flush();
	if (ln0)
		write(1, ln0, (ln1 - 1) - ln0);
	tputs(tgoto(CM, 0, LI - 1), 0, term_putc);
	term_flush();
	rl_on_new_line();
	rl_redisplay();

	return 0;
}

static void
line_up ()
{
	struct screen *scr = scr_front;
	char *ln0, *ln1;
	int li0;

	if (scr->li < (scr->page_lines + 1))
		return;
	li0 = scr->li - (scr->page_lines + 1);
	ln0 = line(scr, li0);
	ln1 = line(scr, li0 + 1);
	if (!ln0 || !ln1)
		return;
	scr->li = li0 + scr->page_lines;
	tputs(tgoto(CM, 0, 0), 0, term_putc);
	tputs(SR, 0, term_putc);
	term_flush();
	write(1, ln0, (ln1 - 1) - ln0);
	tputs(tgoto(CM, rl_point % CO, LI - 1), 0, term_putc);
	term_flush();
}

static void
line_down ()
{
	struct screen *scr = scr_front;
	char *ln0, *ln1;
	int lis = lines(scr->buf, scr->size);

	if (scr->li + 1 > lis)
		return;
	ln0 = line(scr, scr->li);
	if (!ln0)
		return;
	ln1 = line(scr, scr->li + 1);
	if (!ln1)
		ln1 = scr->buf + scr->size;
	scr->li = scr->li + 1;
	if (scr->li > lis) {
		if (scr->li - 1 != lis)
			return;
		scr->li = lis;
	}
	tputs(tgoto(CM, 0, LI - 4), 0, term_putc);
	tputs(SF, 0, term_putc);
	term_flush();
	write(1, ln0, (ln1 - 1) - ln0);
	tputs(tgoto(CM, rl_point % CO, LI - 1), 0, term_putc);
	term_flush();
}

static void
page_up ()
{
	struct screen *scr = scr_front;
	char *ln0, *ln1;
	int li0;

	li0 = (scr->li > (scr->page_lines * 2)) ? (scr->li - (scr->page_lines * 2)) : 0;
	ln0 = line(scr, li0);
	ln1 = line(scr, li0 + scr->page_lines);
	if (!ln0 || !ln1)
		return;
	scr->li = li0 + scr->page_lines;
	draw(ln0, (ln1 - 1) - ln0);
}

static void
page_down ()
{
	struct screen *scr = scr_front;
	char *ln0, *ln1;
	int lis = lines(scr->buf, scr->size);

	if (scr->li + scr->page_lines > lis)
		ln0 = line(scr, lis - scr->page_lines);
	else
		ln0 = line(scr, scr->li);
	if (!ln0)
		return;
	ln1 = line(scr, scr->li + scr->page_lines);
	if (!ln1) {
		ln1 = scr->buf + scr->size;
		scr->li = lis;
	} else
		scr->li = scr->li + scr->page_lines;
	draw(ln0, (ln1 - 1) - ln0);
}

static void
home ()
{
	struct screen *scr = scr_front;
	char *ln0, *ln1;

	ln0 = line(scr, 0);
	ln1 = line(scr, scr->page_lines);
	if (!ln0 || !ln1)
		return;
	if (scr->li == scr->page_lines)
		return;
	scr->li = scr->page_lines;
	draw(ln0, (ln1 - 1) - ln0);
}

static void
end ()
{
	struct screen *scr = scr_front;
	char *ln0, *ln1;
	int lis = lines(scr->buf, scr->size);

	if (lis < scr->page_lines)
		return;
	ln0 = line(scr, lis - scr->page_lines);
	if (!ln0)
		return;
	ln1 = scr->buf + scr->size;
	scr->li = lis;
	draw(ln0, (ln1 - 1) - ln0);
}

void
hx_printf (const char *fmt, ...)
{
	va_list ap;
	va_list save;
	struct screen *scr = scr_front;
	char *buf;
	size_t mal_len;
	int len;

	__va_copy(save, ap);
	mal_len = 256;
	buf = scr->buf;
	for (;;) {
		buf = xrealloc(buf, scr->size + mal_len);
		va_start(ap, fmt);
		len = vsnprintf(buf + scr->size, mal_len, fmt, ap);
		va_end(ap);
		if (len != -1)
			break;
		__va_copy(ap, save);
		mal_len <<= 1;
	}
	scr->buf = buf;
	if (scr->page_lines != LI - 3)
		scr->page_lines = LI - 3;
	if (scr->li == lines(scr->buf, scr->size)) {
		scr->li = lines(scr->buf, scr->size + len);
		term_puts(buf + scr->size, len);
	}
	scr->size += len;
}

static void
term_clear (void)
{
	if (scr_front->buf)
		xfree(scr_front->buf);
	memset(scr_front, 0, sizeof(struct screen));
	redraw();
}

void
hx_save (char *filename)
{
	int f;
	ssize_t r;
	size_t pos, len;
	char path[MAXPATHLEN];

	expand_tilde(path, filename);
	f = open(path, O_WRONLY|O_CREAT|O_APPEND, S_IRUSR|S_IWUSR);
	if (f < 0) {
		hx_printf("save: %s: %s\n", path, strerror(errno));
		return;
	}
	pos = 0;
	len = scr_front->size;
	while (len) {
		r = write(f, scr_front->buf + pos, len);
		if (r <= 0)
			break;
		pos += r;
		len -= r;
	}
	fsync(f);
	close(f);
	hx_printf(INFOPREFIX "%d bytes written to %s\n", pos, path);
}

static void
hotline_client_read_tty (int fd)
{
	char c;

	if (read(fd, &c, 1) <= 0) {
		FD_CLR(fd, &hxd_rfds);
		return;
	}
	rl_pending_input = c;
	rl_callback_read_char();
}

static void
term_rl_input (void)
{
	hotline_client_input(rl_line_buffer);
}

extern int expand_command (char *, int, char *);

static void
last_msg ()
{
	char buf[128], nick[128];
	int len, srlp;

	if (!*last_msg_nick)
		return;

	strunexpand(last_msg_nick, strlen(last_msg_nick), nick, sizeof(nick));
	len = sprintf(buf, "/msg %.120s ", nick);
	srlp = rl_point;
	rl_point = 0;
	rl_insert_text(buf);
	rl_point = len + srlp;
}

static int
tab ()
{
	if (*rl_line_buffer == '/') {
		char cmdbuf[16], *cmd = rl_line_buffer + 1;
		int i, l, srlp;

		for (l = 0; cmd[l]; l++)
			if (cmd[l] == ' ')
				break;
		i = expand_command(cmd, l, 1+cmdbuf);
		if (i >= 0 && ((int)strlen(1+cmdbuf) != l || !cmd[l])) {
			srlp = rl_point;
			cmdbuf[0] = '/';
			if (!i) {
				i = strlen(cmdbuf);
				cmdbuf[i++] = ' ';
				cmdbuf[i] = 0;
				if (cmd[l] && cmd[l+1])
					l++;
			} else
				i = -1;
			rl_point = 0;
			rl_delete_text(0, 1+l);
			rl_insert_text(cmdbuf);
			rl_point = i < 0 ? srlp : srlp + ((i - 1) - l);
		}
		cmd = rl_line_buffer + 1;
		if (!(cmd_arg(1, cmd) & 0xffff))
			return 0;
		if (!strncmp(cmd, "cd", 2)
		    || !strncmp(cmd, "ln", 2)
		    || !strncmp(cmd, "ls", 2)
		    || !strncmp(cmd, "mv", 2)
		    || !strncmp(cmd, "rm", 2)
		    || (!strncmp(cmd, "get", 3) && cmd[3] == ' ')
		    || !strncmp(cmd, "finfo", 5)) {
			int argn, ambig;
			u_int16_t off, len, orig_len;
			u_int32_t ol;
			char path[MAXPATHLEN], xxx[0x1000];

			argn = 1;
			while ((ol = cmd_arg(argn++, cmd))) {
				off = ol >> 16, orig_len = ol & 0xffff;
				if ((rl_point-1) >= off && (rl_point-1) <= off+orig_len)
					goto ok;
			}
			return 0;
ok:
			orig_len &= 0x0fff;
			memcpy(xxx, cmd+off, orig_len);
			len = strexpand(xxx, orig_len);
			ambig = expand_path(xxx, len, path);
			if (ambig >= 0) {
				char *p, buf[MAXPATHLEN * 3 + 8];

				p = buf;
				len = strunexpand(path, strlen(path), p, sizeof(buf) - (p - buf));
				if (!ambig) {
					p += len;
					*p++ = ' ';
					*p = 0;
					len++;
				}
				rl_delete_text(1 + off, 1 + off + orig_len);
				rl_point = 1 + off;
				rl_insert_text(buf);
				srlp = rl_point;
				rl_point = 0;
				hx_output.clear_input();
				rl_point = srlp;
				rl_on_new_line();
				rl_redisplay();
			}
		} else if (!strncmp(cmd, "msg", 3)
			 || !strncmp(cmd, "info", 4)
			 || !strncmp(cmd, "kick", 4)
			 || !strncmp(cmd, "chat", 4)
			 || !strncmp(cmd, "invite", 3)) {
			u_int32_t ol = cmd_arg(1, cmd);
			int off = ol >> 16, len = ol & 0xffff;
			char *n, xxx[128];

			if (!off)
				return 0;
			len &= 0x7f;
			memcpy(xxx, cmd+off, len);
			len = strexpand(xxx, len);
			n = expand_user_name(xxx, len);
			if (n && *(n+len)) {
				char *p, buf[120];

				p = buf + sprintf(buf, "/%.*s ", off-1, cmd);
				len = strunexpand(n, strlen(n), p, sizeof(buf) - (p - buf));
				p += strlen(p);
				*p++ = ' ';
				*p = 0;
				rl_point = 0;
				rl_delete_text(0, 1+off+len);
				rl_insert_text(buf);
			}
		} else if (!strncmp(cmd, "exec", 4)
			   || !strncmp(cmd, "lcd", 3)
			   || !strncmp(cmd, "load", 4)
			   || !strncmp(cmd, "put", 3)
			   || !strncmp(cmd, "type", 4)) {
			char *buf;
			int len, srlp, erlp, xxx = 0;
			srlp = rl_point;
			rl_complete_internal(TAB);
			erlp = rl_point;
			if (rl_line_buffer[erlp-1] == ' ' && (erlp != srlp))
				xxx = 1;
			len = erlp-srlp-xxx;
			if (len) {
				buf = xmalloc(len*3);
				len = strunexpand(rl_line_buffer+srlp, len, buf, len*3);
				if (xxx) {
					buf[len++] = ' ';
					buf[len] = 0;
				}
				rl_delete_text(srlp, erlp);
				rl_point = srlp;
				rl_insert_text(buf);
				xfree(buf);
			}
		}
	} else if (!rl_end)
		last_msg();

	return 0;
}

static void
init (int argc, char **argv)
{
	int fd;
	struct opt_r opt;
	int i;

	opt.err_printf = hxd_log;
	opt.ind = 0;
	while ((i = getopt_r(argc, argv, "d", &opt)) != EOF) {
		switch (i) {
			case 'd':
				dumb = 1;
				break;
		}
	}

	fd = term_init();
	if (fd == -1)
		return;
	if (fd >= hxd_open_max) {
		hxd_log("%s:%d: %d >= hxd_open_max (%d)", __FILE__, __LINE__, fd, hxd_open_max);
		return;
	}
	hxd_files[fd].ready_read = hotline_client_read_tty;
	FD_SET(fd, &hxd_rfds);

	rl_bind_key('\t', tab);
	rl_generic_bind(ISFUNC, "\033\t", (char *)last_msg, rl_get_keymap());
	rl_readline_name = "hx";
	rl_callback_handler_install("", term_rl_input);
}

static void
cleanup ()
{
	term_reset();
}

static void
clear_input (void)
{
	term_goto_line(LI - 1);
	term_ce();
}

static void
output_chat (u_int32_t cid, char *chat, u_int16_t chatlen)
{
	if (cid)
		hx_printf("0x%08x: %.*s\n", cid, chatlen, chat);
	else
		hx_printf("%.*s\n", chatlen, chat);
}

static void
output_msg (u_int32_t uid, char *name, char *msgbuf, u_int16_t msglen)
{
	char *col;
	struct user *user;

	user = user_with_uid(chat_list->user_list, uid);
	col = user ? (user->color % 4 > 1 ? (user->color % 2 ? RED : RED_BOLD)
					  : (user->color % 2 ? GREEN : GREEN_BOLD))
		   : GREEN;
	hx_printf("[%s%s%s(%u)]: %.*s\n", col, name, DEFAULT, uid, msglen, msgbuf);
}

static void
output_agreement (char *agreement, u_int16_t len)
{
	hx_printf(INFOPREFIX "Agreement:\n%.*s\n", len, agreement);
}

struct output_functions hx_output = {
	init,
	cleanup,
	term_status,
	term_clear,
	clear_input,
	term_mode_underline,
	term_mode_clear,
	output_chat,
	output_msg,
	output_agreement
};
